//
//  CardStackView.m
//  Card Game
//
//  Created by Amin Negm-Awad
//  Copyright 2009 [co coa:ding]. All rights reserved.

#import "CardStackView.h"
#import "TableBaizeView.h"

#import "BezierPathCardsAddition.h"
#import "KeyValueBindingExtension.h"

// Kartenbemassung
static const float cardSizeRatio = 1.6;
static const float edgeRadius = 0.1;
static const NSSize stepSize = { 0.21, 0.11 };

// Positionen in Kartengroesse
static const NSPoint labelLocation = { 0.05, 0.9 };
static const CGFloat labelHeight   = 0.08;
@interface CardStackView( PrivateMethodsAddition )
- (NSRect)rectForCardWithIndex:(NSUInteger)index 
                         count:(NSUInteger)count
                   getIsJammed:(BOOL*)jammed;
@end

NSString* const disclosedCardsBinding 
= @"disclosedCardsBinding";
NSString* const nonDisclosedCardsCountBinding
= @"nonDisclosedCardsCountBinding";
static NSString* const suitBinding = @"suitBinding";
static NSString* const valueBinding = @"valueBinding";

@implementation CardStackView
@synthesize alignment, draggedCardsCount;
@synthesize valueKeyPath, suitKeyPath;

- (void)   setInfo:(NSDictionary*)info 
        forBinding:(NSString*)binding
{
   [bindingInfos setObject:info forKey:binding];
}

- (void)removeInfoForBinding:(NSString*)binding
{
   [bindingInfos removeObjectForKey:binding];
}

- (NSDictionary*)infoForBinding:(NSString*)binding
{
   NSDictionary* info = [bindingInfos objectForKey:binding];
   if( info ) {
      return info;
   }
   
   return [super infoForBinding:binding];
}

- (NSDictionary*)bindingInfos 
{ 
   return [[bindingInfos copy] autorelease];
}

- (void)setBindingInfos:(NSDictionary*)value 
{
   if( value != bindingInfos ) {
      [bindingInfos release];
      bindingInfos = [value mutableCopy];
   }
}

- (NSUInteger)countOfNonDisclosedCards 
{ 
   return countOfNonDisclosedCards; 
}
- (void)setCountOfNonDisclosedCards:(NSUInteger)value 
{
   countOfNonDisclosedCards = value;
   [self setNeedsDisplay:YES];
}

- (NSUInteger)cardsCount 
{
   return [self.cards count] + self.countOfNonDisclosedCards;
}

- (void)setSelectedCardsIndices:(NSIndexSet*)value {
   if( value != selectedCardsIndices ) {
      [selectedCardsIndices release];
      selectedCardsIndices = [value retain];
   }
   
   [self setNeedsDisplay:YES];
}

- (NSIndexSet*)selectedCardsIndices 
{ 
   return selectedCardsIndices;
}

- (NSArray*)cards { return cards; }
- (void)setCards:(NSArray*)value
{
   // Aenderungsrechtecke herausfinden
   NSRect dirtyRect;
   
   // Die Anzahl der vorhandene Karten und Stauchung holen 
   NSUInteger nonDisclosedCount = self.countOfNonDisclosedCards;
   NSUInteger oldCount = [cards count] + nonDisclosedCount;
   NSUInteger newCount = [value count] + nonDisclosedCount;
   
   BOOL oldJammed;
   BOOL newJammed;
   [self rectForCardWithIndex:0 
                        count:oldCount
                  getIsJammed:&oldJammed];
   [self rectForCardWithIndex:0 
                        count:newCount
                  getIsJammed:&newJammed];
   
   // Wenn mindestens ein Stapel gestaucht ist und sich die An-
   // zahl der Karten veraendert, muss ohnehin das gesamte View
   // neu gezeichnet werden
   if( (oldJammed || newJammed) && (oldCount != newCount ) ) {
      [self setNeedsDisplay:YES];
      
      // Andernfalls wissen wir hier, dass entweder keine Stauchung 
      // vorliegt, der count also für die Berechnung der Rechtecke 
      // unbeachtlich ist, oder aber beide Stapel dieselbe Anzahl
      // von Karten enthalten, so dass beide counts gleich sind.
   } else {
      // Anzahl der gemeinsamen Karten
      NSUInteger count = (oldCount < newCount)?oldCount:newCount;
      NSUInteger index;
      for( index = nonDisclosedCount; index < count; index++ ) {
         NSUInteger cardIndex = index - nonDisclosedCount;
         id oldCard = [cards objectAtIndex:cardIndex];
         id newCard = [value objectAtIndex:cardIndex];
         
         // Unterschiedlicher Wert
         NSNumber* oldValue = [oldCard valueForKey:@"value"];
         NSNumber* newValue = [newCard valueForKey:@"value"];
         if( ![oldValue isEqualToNumber:newValue] ) {
            dirtyRect = [self rectForCardWithIndex:index
                                             count:count
                                       getIsJammed:&oldJammed];
            [self setNeedsDisplayInRect:dirtyRect]; 
            continue;
            //////////
         }
         
         // Unterschiedliche Farbe
         NSNumber* oldSuit = [oldCard valueForKey:@"suit"];
         NSNumber* newSuit = [newCard valueForKey:@"suit"];
         if( ![oldSuit isEqualToNumber:newSuit] ) {
            dirtyRect = [self rectForCardWithIndex:index
                                             count:count
                                       getIsJammed:&oldJammed];
            [self setNeedsDisplayInRect:dirtyRect];
            continue;
            //////////
         }
      }
      
      // Unterschiedliche Stapelhoehen beruecksichtigen:
      // index zeigt bisher auf erste "ueberstehende" Karte des 
      // groesseren Stapels
      
      // Letzte Karte holen
      count  = (oldCount > newCount)?oldCount:newCount;
      
      // Den Bereich zwischen den Karten markieren
      NSRect lastRect;
      dirtyRect = [self rectForCardWithIndex:index
                                       count:count
                                 getIsJammed:&oldJammed];
      lastRect  =  [self rectForCardWithIndex:count-1
                                        count:count
                                  getIsJammed:&oldJammed];
      
      // Der gesamte Bereich muss neu gezeichnet werden. 
      dirtyRect = NSUnionRect( dirtyRect, lastRect );
      
      [self setNeedsDisplayInRect:dirtyRect];   
   }
   
   // Standardaccessor
   if( value != cards ) {
      [cards release];
      cards = [value mutableCopy];
   }
}

- (void)     insertCards:(NSArray*)value 
        inCardsAtIndexex:(NSIndexSet*)indexes
{
   [cards insertObjects:value atIndexes:indexes];
   [self setNeedsDisplay:YES];
}

- (void)removeCardsAtIndexes:(NSIndexSet*)indexes
{
   [cards removeObjectsAtIndexes:indexes];
   [self setNeedsDisplay:YES];
}

- (void)replaceCardsAtIndexes:(NSIndexSet*)indexes
                    withCards:(NSArray*)value
{
   [cards replaceObjectsAtIndexes:indexes withObjects:value];
   [self setNeedsDisplay:YES];
}
#pragma mark -

#pragma mark Drawing
- (void)drawUpperLabelWithSuit:(NSUInteger)suit
                         value:(NSUInteger)value
                        inRect:(NSRect)rect
{
   NSBezierPath* path;
   
   // Berechne das Label-Rect
   NSPoint location;
   location.x 
   = NSMinX( rect ) + NSWidth( rect )  * labelLocation.x; 
   location.y 
   = NSMinY( rect ) + NSHeight( rect ) * labelLocation.y; 
   CGFloat height = rect.size.height * labelHeight;
   
   // Farbsymbol erzeugen, anpassen und ausgeben
   path = [NSBezierPath bezierPath];
   [path moveToPoint:NSMakePoint( 0.0, 0.0)];
   [path appendBezierPathForSuit:suit];
   
   [path transformToLocation:location height:height];
   
   [path fill];
   location.x = NSMaxX( [path bounds] ) + 0.01 * NSWidth( rect );
   
   // Wertesymbol erzeugen, anpassen und ausgeben
   path = [NSBezierPath bezierPath];
   [path moveToPoint:NSMakePoint( 0.0, 0.0)];
   [path appendBezierPathForValue:value];
   
   [path transformToLocation:location height:height];
   
   [path fill];
}

- (void)drawLowerLabelWithSuit:(NSUInteger)suit
                         value:(NSUInteger)value
                        inRect:(NSRect)rect
{
   NSBezierPath* path;
   
   // Berechne das Label-Rect
   NSPoint location;
   location.x 
      = NSMaxX( rect ) - NSWidth( rect )  * labelLocation.x; 
   location.y 
      = NSMaxY( rect ) - NSHeight( rect ) * labelLocation.y; 
   CGFloat height = -1.0 * rect.size.height * labelHeight;
   
   // Farbsymbol erzeugen, anpassen und ausgeben
   path = [NSBezierPath bezierPath];
   [path moveToPoint:NSMakePoint( 0.0, 0.0)];
   [path appendBezierPathForSuit:suit];
   
   [path transformToLocation:location height:height];
   
   [path fill];
   location.x = NSMinX( [path bounds] ) - 0.01 * NSWidth( rect );
   
   // Wertesymbol erzeugen, anpassen und ausgeben
   path = [NSBezierPath bezierPath];
   [path moveToPoint:NSMakePoint( 0.0, 0.0)];
   [path appendBezierPathForValue:value];
   
   [path transformToLocation:location height:height];
   
   [path fill];
}

- (void)drawCardFaceWithSuit:(NSUInteger)suit
                       value:(NSUInteger)value
                      inRect:(NSRect)rect 
{
   if( suit < 2 ) {
      [[NSColor redColor] set];
   } else {
      [[NSColor blackColor] set];
   }
   
   [self drawUpperLabelWithSuit:suit value:value inRect:rect];
   [self drawLowerLabelWithSuit:suit value:value inRect:rect];
}

- (void)drawCardBorderInRect:(NSRect)rect selected:(BOOL)selected
{
   NSBezierPath* path;
   
   CGFloat radius = NSWidth( rect ) * edgeRadius;
   
   path = [NSBezierPath bezierPathWithRoundedRect:rect 
                                          xRadius:radius 
                                          yRadius:radius];
   [path setLineWidth:0.0];
   
   // Selektierte Karten:
   if( selected ) {
      [[NSColor selectedControlColor] setFill];
   } else {
      [[NSColor whiteColor] setFill];
   }
   [path fill];
   
   [[NSColor blackColor] setStroke];
   [path stroke];
}
+ (CGFloat)cardHeightForWidth:(CGFloat)width {
   return width * cardSizeRatio;
}

+ (CGFloat)cardWidthForHeight:(CGFloat)height {
   return height / cardSizeRatio;
}

- (NSRect)rectForCardWithIndex:(NSUInteger)index 
                         count:(NSUInteger)count
                   getIsJammed:(BOOL*)isJammed
{
   CGFloat xStep;
   CGFloat yStep;
   
   NSRect rect = [self bounds];
   
   // Startwert und Schrittweite nach Ausrichtungen setzen
   switch( self.alignment ) {
      case NSMinXEdge: // linksbuendig
         rect.size.width 
         = [[self class] cardWidthForHeight:NSHeight( rect )];
         rect.origin.x = 0.0;
         xStep = NSWidth( rect ) * +stepSize.width;
         yStep = 0.0;
         break;
         
      case NSMaxXEdge: // rechtbuendig
         rect.size.width 
         = [[self class] cardWidthForHeight:NSHeight( rect )];
         rect.origin.x 
         = NSMaxX( [self bounds] ) - NSWidth( rect );
         xStep = NSWidth( rect ) * -stepSize.width;
         yStep = 0.0;
         break;
         
      case NSMinYEdge: // untenbuendig (hier nicht verwendet)
         rect.size.height
         = [[self class] cardHeightForWidth:NSWidth( rect )];
         rect.origin.y = 0.0;
         xStep = 0.0;
         yStep = NSHeight( rect ) * +stepSize.height;
         break;
         
      case NSMaxYEdge: // obenbuendig
         rect.size.height
         = [[self class] cardHeightForWidth:NSWidth( rect )];
         rect.origin.y 
         = NSMaxY( [self bounds] ) - NSHeight( rect );
         xStep = 0.0;
         yStep = NSHeight( rect ) * -stepSize.height;
         
         break;
      default:
         break;
   }
   
   // Ueberlaeufe abfangen
   BOOL jammed = NO;
   CGFloat maxStep;
   
   count--;
   if( count > 0 ) {
      maxStep = (NSHeight([self bounds])-NSHeight(rect)) / count;
      if( yStep < 0.0 ) {
         if( -yStep > maxStep ) {
            yStep = -maxStep;
            jammed = YES;
         }
      } else {
         if( yStep > maxStep ) {
            yStep = maxStep;
            jammed = YES;
         }
      }
      
      maxStep = (NSWidth([self bounds])-NSWidth(rect)) / count;
      if( xStep < 0.0 ) {
         if( -xStep > maxStep ) {
            xStep = -maxStep;
            jammed = YES;
         } 
      } else {
         if( xStep > maxStep ) {
            xStep = +maxStep;
            jammed = YES;
         } 
      }         
   }
   if( isJammed != NULL ) {
      *isJammed = jammed;
   }
   
   
   // Verschiebe Karte um Index
   rect = NSOffsetRect( rect, xStep * index, yStep * index );
   
   // Pixelmaessig ausrichten
   rect = [self convertRectToBase:rect]; 
   rect.origin.x = ceilf( rect.origin.x ) + 0.5;
   rect.origin.y = ceilf( rect.origin.y ) + 0.5;
   rect.size.width =floorf( rect.size.width ) - 2.0;
   rect.size.height =floorf( rect.size.height ) - 2.0;
   rect = [self convertRectFromBase:rect]; 
   
   return rect;      
}
- (void)drawCardBackInRect:(NSRect)rect
{
   NSBezierPath* path;
   
   CGFloat radius = NSWidth( rect ) * edgeRadius;
   
   path = [NSBezierPath bezierPathWithRoundedRect:rect 
                                          xRadius:radius 
                                          yRadius:radius];
   [[NSColor lightGrayColor] setFill];
   [path setLineWidth:0.0];
   [path fill];
   
   [[NSColor blackColor] setStroke];
   [path stroke];
}

- (void)drawRect:(NSRect)rect {
   NSUInteger suit;
   NSUInteger value;
   NSUInteger index;
   
   // Zeichne verdeckte Karten
   NSUInteger nonDisclosedCount = self.countOfNonDisclosedCards;
   NSUInteger count = self.cardsCount;
   
   // Falls zugedeckte Karten gezogen wurde (eigentlich nie)
   if( [self.cards count] < self.draggedCardsCount ) {
      // muessen diese abgezogen werden
      count -= self.draggedCardsCount - [self.cards count];
   }
   
   for( index = 0; index < nonDisclosedCount; index++ ) { 
      NSRect cardRect;
      cardRect = [self rectForCardWithIndex:index 
                                      count:count 
                                getIsJammed:NULL];
      [self drawCardBackInRect:cardRect];
   }
   
   
   for( ; index < (count - self.draggedCardsCount); index++ ) { 
      NSUInteger arrayIndex = index - nonDisclosedCount;
	   id card = [self.cards objectAtIndex:arrayIndex];
      
      NSRect cardRect;
      cardRect = [self rectForCardWithIndex:index 
                                      count:count 
                                getIsJammed:NULL];
      
      if( [self needsToDrawRect:cardRect] ) {
         BOOL selected 
            = [self.selectedCardsIndices containsIndex:index];
         [self drawCardBorderInRect:cardRect selected:selected];
         suit = [[card valueForKey:@"suit"] integerValue];
         value = [[card valueForKey:@"value"] integerValue];
         [self drawCardFaceWithSuit:suit 
                              value:value 
                             inRect:cardRect];
      }
   }    
}
#pragma mark -

#pragma mark Events
- (NSInteger)indexOfCardAtLocation:(NSPoint)location
{
   NSUInteger cardCount = self.cardsCount;;
   NSInteger cardIndex;
   
   for( cardIndex = cardCount-1; cardIndex > -1; cardIndex-- ) {
      NSRect cardRect = [self rectForCardWithIndex:cardIndex 
                                             count:cardCount
                                       getIsJammed:NULL];
      
      if( NSPointInRect( location, cardRect ) ) {
         // gefunden
         break;
      }
   }
   
   return cardIndex;
}

- (void)mouseDragged:(NSEvent*)event 
{
   // Bitte gehen Sie weiter, hier gibt es nichts zu sehen
   NSUInteger count = self.cardsCount;
   if( count == 0) {
      return;
   }
   
   // Image erzeugen
   //    gezogene Karte besorgen
   NSPoint location = [event locationInWindow];
   location = [self convertPoint:location fromView:nil];
   NSInteger index = [self indexOfCardAtLocation:location];
   
   //    Falls ausserhalb des Stapels: Weg!
   if( index == -1 ) {
      return;
   }
   
   // Falls zugedeckte Karte: Weg!
   if( index < self.countOfNonDisclosedCards ) {
      return;
   }

   // Auch der TableBaizeView muss das Dragging erlauben.
   // Dies befragt wiederum sein Delegate.
   TableBaizeView* tableBaizeView = (TableBaizeView*)[self superview];
   BOOL allowed
   = [tableBaizeView allowsDraggingFromCardStackView:self    
                                             atIndex:index];
   if( !allowed ) {
      return;
   }
   
   //    Groesse berechnen
   NSRect draggedRect = [self rectForCardWithIndex:index
                                             count:count
                                       getIsJammed:NULL];
   NSRect lastRect = [self rectForCardWithIndex:count-1
                                          count:count
                                    getIsJammed:NULL];
   NSRect imageRect = NSUnionRect( draggedRect, lastRect );
   NSSize offset = NSMakeSize( imageRect.origin.x - 10.5, 
                              imageRect.origin.y -10.5 );
   imageRect.size.width  += 21.0;
   imageRect.size.height += 21.0;
   
   //    Bild zeichnen
   NSImage* dragImage = [[[NSImage alloc] 
                          initWithSize:imageRect.size] 
                         autorelease];
   [dragImage lockFocus];

   
   NSInteger suit;
   NSInteger value;
   NSInteger drawIndex;
   NSUInteger cardIndex;
   for( drawIndex = index; drawIndex < count; drawIndex++ ) { 
      cardIndex = drawIndex - self.countOfNonDisclosedCards;
      id card = [self.cards objectAtIndex:cardIndex];
      
      NSRect drawRect = [self rectForCardWithIndex:drawIndex 
                                             count:count
                                       getIsJammed:NULL];
      drawRect = NSOffsetRect( drawRect, -offset.width, 
                              -offset.height );
      
      [NSGraphicsContext saveGraphicsState];
      NSShadow* shadow = [[[NSShadow alloc] init] autorelease];
      [shadow setShadowBlurRadius:10.0];
      [shadow set];
      [self drawCardBorderInRect:drawRect selected:NO];
      [NSGraphicsContext restoreGraphicsState];
      
      suit = [[card valueForKey:@"suit"] integerValue];
      value = [[card valueForKey:@"value"] integerValue];
      [self drawCardFaceWithSuit:suit 
                           value:value 
                          inRect:drawRect];
      
   } 
   [dragImage unlockFocus];
   
   // Pasteboard erstellen und befuellen
   NSPasteboard* pasteboard 
   = [NSPasteboard pasteboardWithName:NSDragPboard];

   NSArray* dragTypes 
   = [NSArray arrayWithObject:NSTIFFPboardType];
   [pasteboard declareTypes:dragTypes owner:nil];
   [pasteboard setData:[dragImage TIFFRepresentation] 
               forType:NSTIFFPboardType];  
   
   // Gezogene Karten ausblenden
   self.draggedCardsCount = self.cardsCount - index;
   [self setNeedsDisplay:YES];
   
   // Ab dafuer
   imageRect.origin.x -= 10.5;
   imageRect.origin.y -= 10.5;
   [self dragImage:dragImage 
                at:imageRect.origin offset:NSZeroSize
             event:event
        pasteboard:pasteboard
            source:self
         slideBack:YES];

   // Gezogene Karten wieder einfuegen
   self.draggedCardsCount = 0;
   [self setNeedsDisplay:YES];
   
}

#pragma mark -

#pragma mark Dragging Source
- (NSDragOperation)draggingSourceOperationMaskForLocal:(BOOL)local {
   if( local ) {
      return   NSDragOperationMove 
      | NSDragOperationCopy 
      | NSDragOperationDelete;
   } else {
      return NSDragOperationCopy;
   }
}

- (BOOL)ignoreModifierKeysWhileDragging { return YES; }
#pragma mark -

#pragma mark Dragging Destination
- (NSDragOperation)draggingEntered:(id<NSDraggingInfo>)info
{
   NSDragOperation op = [info draggingSourceOperationMask];
   
   if( self.cardsCount  > 0 ) {
      self.selectedCardsIndices 
         = [NSIndexSet indexSetWithIndex:self.cardsCount - 1];
      [self setNeedsDisplay:YES];
   }
   
   if( op & NSDragOperationMove ) {
      return NSDragOperationMove;
   } else if( op & NSDragOperationCopy ) {
      return NSDragOperationCopy;
   } else {
      return NSDragOperationNone;
   }
}

- (void)draggingExited:(id<NSDraggingInfo>)info
{
   self.selectedCardsIndices = [NSIndexSet indexSet];
   [self setNeedsDisplay:YES];
}

- (void)draggingEnded:(id<NSDraggingInfo>)info
{
   self.selectedCardsIndices = [NSIndexSet indexSet];
}

- (BOOL)wantsPeriodicDraggingUpdates  { return NO; }

- (NSDragOperation)draggingUpdated:(id<NSDraggingInfo>)info
{
   NSDragOperation op = [info draggingSourceOperationMask];
   if( op & NSDragOperationMove ) {
      return NSDragOperationMove;
   } else if( op & NSDragOperationCopy ) {
      return NSDragOperationCopy;
   } else {
      return NSDragOperationNone;
   }
}

- (BOOL)performDragOperation:(id<NSDraggingInfo>)info
{
   return NO;
}
#pragma mark -

#pragma mark Bindings
- (void)removeObserverFromObjects:(NSArray*)items
{
   NSRange oldRange = NSMakeRange( 0, [items count] );
   NSIndexSet* indexes;
   indexes = [NSIndexSet indexSetWithIndexesInRange:oldRange];            
   [items        removeObserver:self 
           fromObjectsAtIndexes:indexes 
                     forKeyPath:self.valueKeyPath];
   [items        removeObserver:self 
           fromObjectsAtIndexes:indexes 
                     forKeyPath:self.suitKeyPath];
}

- (void)addObserverToObjects:(NSArray*)items
{
   NSRange oldRange = NSMakeRange( 0, [items count] );
   NSIndexSet* indexes;
   indexes = [NSIndexSet indexSetWithIndexesInRange:oldRange];            
   
   [items        addObserver:self 
          toObjectsAtIndexes:indexes
                  forKeyPath:self.valueKeyPath 
                     options:0
                     context:valueBinding];
   [items        addObserver:self 
          toObjectsAtIndexes:indexes
                  forKeyPath:self.suitKeyPath 
                     options:0
                     context:suitBinding];
}

- (void)observeValueForKeyPath:(NSString*)keyPath 
                      ofObject:observed 
                        change:(NSDictionary*)change 
                       context:(id)context
{
   // Ist es meine Observierung?
   if( context == disclosedCardsBinding ) {
      NSDictionary* info;
      info = [self infoForBinding:disclosedCardsBinding];
      NSDictionary* options;
      options = [info objectForKey:NSOptionsKey];
      NSArray* new = [observed valueForKeyPath:keyPath];
      NSValueTransformer* transformer
      = [options objectForKey:NSValueTransformerBindingOption];
      if( transformer ) {
         new = [transformer transformedValue:new];
      }
      NSArray* old = self.cards;

      [self removeObserverFromObjects:old];
      self.cards = new;
      [self addObserverToObjects:new];
   
   } else if( context == suitBinding  ) {
      [self setNeedsDisplay:YES];
      
   } else if( context == valueBinding  ) {
      [self setNeedsDisplay:YES];      
      
   } else if( context == nonDisclosedCardsCountBinding ) {
      NSDictionary* info;
      info = [self infoForBinding:nonDisclosedCardsCountBinding];
      NSNumber* new = [observed valueForKeyPath:keyPath];
      self.countOfNonDisclosedCards 
      = [new unsignedIntegerValue];
   
   } else {
      [super observeValueForKeyPath:keyPath 
                           ofObject:observed
                             change:change
                            context:context];
   }
}

- (void)       bind:(NSString*)binding 
           toObject:(id)observed 
        withKeyPath:(NSString*)keyPath 
            options:(NSDictionary*)bindingOptions
{
   if( binding == disclosedCardsBinding ) {
      if( !bindingOptions ) {
         bindingOptions = [NSDictionary dictionary];
      }
      NSDictionary* info;
      info = [NSDictionary dictionaryWithObjectsAndKeys:
              observed,        NSObservedObjectKey,
              keyPath,         NSObservedKeyPathKey,
              bindingOptions,  NSOptionsKey,
              nil];
      [self setInfo:info forBinding:disclosedCardsBinding];
      
      NSKeyValueObservingOptions options;
      options =   NSKeyValueObservingOptionNew
      | NSKeyValueObservingOptionOld
      | NSKeyValueObservingOptionInitial;
      
      [observed addObserver:self
                 forKeyPath:keyPath 
                    options:options
                    context:binding];
      
   } else if( binding == nonDisclosedCardsCountBinding ) {
      if( !bindingOptions ) {
         bindingOptions = [NSDictionary dictionary];
      }
      NSDictionary* info;
      info = [NSDictionary dictionaryWithObjectsAndKeys:
              observed,        NSObservedObjectKey,
              keyPath,         NSObservedKeyPathKey,
              bindingOptions,  NSOptionsKey,
              nil];
      
      [self    setInfo:info 
            forBinding:nonDisclosedCardsCountBinding];
      
      NSKeyValueObservingOptions options;
      options =   NSKeyValueObservingOptionNew
      | NSKeyValueObservingOptionInitial;
      
      [observed addObserver:self
                 forKeyPath:keyPath 
                    options:options
                    context:nonDisclosedCardsCountBinding];    
            
   } else {
      [super        bind:binding 
                toObject:observed 
             withKeyPath: keyPath
                 options:bindingOptions];
   }
}

- (void)unbind:(NSString*)binding
{
   if( binding == disclosedCardsBinding ) {
      // Abhaengige Observierungen loeschen.
      [self removeObserverFromObjects:cards];
      NSDictionary* info 
         = [self infoForBinding:disclosedCardsBinding];
      id observed = [info objectForKey:NSObservedObjectKey];
      NSString* keyPath =
      [info objectForKey:NSObservedKeyPathKey];
      [observed removeObserver:self forKeyPath:keyPath];
      
   } else if( binding == nonDisclosedCardsCountBinding ) {
      NSDictionary* info 
      = [self infoForBinding:disclosedCardsBinding];
      id observed = [info objectForKey:NSObservedObjectKey];
      NSString* keyPath =
      [info objectForKey:NSObservedKeyPathKey];
      [observed removeObserver:self forKeyPath:keyPath];
      [self removeInfoForBinding:binding];
      
      
   } else {
      [super unbind:binding];
   }
}
#pragma mark -

#pragma mark Delegating

#pragma mark -

#pragma mark Instantiation
- (id)initWithFrame:(NSRect)frame {
   self = [super initWithFrame:frame];
   if (self) {
      self.bindingInfos = [NSDictionary dictionary];
      self.alignment = NSMaxYEdge;
      self.cards = [NSArray array];
      self.alignment = NSMaxXEdge;
      self.selectedCardsIndices = [NSIndexSet indexSet];
   }
   return self;
}

- (void) dealloc
{
   self.bindingInfos = nil;
   self.selectedCardsIndices = nil;
   self.cards = nil;
   self.alignment = 0;
   self.draggedCardsCount = 0;
   [super dealloc];
}
@end
